'''
Lunar Lander Cheez Grabber!

Programmed by Chad Serrant
Art by Parakash Pratibhu

For Lunar Game Jam 2010 (Oct 9 to Oct 10)
'''

import pygame, pygame.font, pygame.gfxdraw
from pygame.locals import *
import time, random, math

class Session():
    def __init__(self):
        self.screen=None
        self.clock=None
        self.quitProgram=False
        self.resetProgram=False
        self.playArea = pygame.Rect(0,0,800,550)
        self.ll = LunarLander()
        self.gravityAccel=0.2
        self.deathTime=None
        self.deathMode="None"
        self.defaultFont=None
        self.highScore=0
        self.terrain=[]
        self.cheezes=[]
        self.aliens=[]
        self.tutorialScreen="Title"
        self.tutorialTime=None
        self.tutorialScroll=800
        #"base" "hover" "wall" "slopeup" "slopedown"
        #"cheez" "alien"
        self.levelGeneratorWeights=[100,50,20,20,5]
        self.levelGeneratorWeightsScreen=[0,0,0,0,0]
        self.levelGeneratorWeightsTemp=[0,0,0,0,0]
        self.cheezGenerator=[0,0,0,0]
        self.alienGenerator=[0,0,0,0]
        self.distanceScrolled=0
        self.lastScrollCheckPoint=800
        self.spriteGroup = pygame.sprite.Group()
        self.lunarLanderImage=None
        self.alien1Image=None
        self.alien2Image=None
        self.alien3Image=None
        self.cheezYellowImage=None
        self.cheezBlueImage=None
        self.cheezRedImage=None
        self.mechClawImageReady=None
        self.mechClawImageRecover=None
        self.mechClawImageLockedOn=None
        self.terrainCursor=[800,550]
        self.backgroundImage=None
        self.backgroundScrollPos=0
        self.backgroundScrollStartTime=None
        self.backgroundScrollDuration=5*60
        self.lunarKidImage=None
        self.alienAudio=None
        self.crashAudio=None
        self.cheezAudio=None
        self.channelLimit=8
    def reset(self):
        self.terrain=[]
        self.cheezes=[]
        self.aliens=[]
        self.playArea = pygame.Rect(0,0,800,550)
        self.ll = LunarLander()
        self.gravityAccel=0.2
        self.terrain=[]
        self.deathTime=None
        self.deathMode="None"
        self.tutorialScreen="Title"
        self.tutorialTime=None
        self.tutorialScroll=800
        self.levelGeneratorWeights=[100,50,10,20,5]
        self.levelGeneratorWeightsScreen=[0,0,0,0,0]
        self.levelGeneratorWeightsTemp=[0,0,0,0,0]
        self.cheezGenerator=[20,33,33,33]
        self.alienGenerator=[15,33,33,33]
        self.distanceScrolled=0
        self.lastScrollCheckPoint=800
        self.spriteGroup.empty()
        self.terrainCursor=[800,550]
        self.backgroundScrollPos=0
        self.backgroundScrollStartTime=None
        self.backgroundImage=pygame.image.load("resources/MoonTerrain.png").convert_alpha()

class Alien(pygame.sprite.Sprite):
    def __init__(self, args=None):
        pygame.sprite.Sprite.__init__(self)
        self.image=None
        self.rect=pygame.Rect(0,0,0,0)
        self.id="alien 1"
        self.moveCount=[0,0]
        self.moveDirection="right"
        self.justOnScreen=False
        if args:
            self.id=args[0]
            self.rect.x=args[1]
            self.rect.y=args[2]
        self.load()
    def load(self):
        if self.id=="alien 1":
            self.image=objs.alien1Image
        if self.id=="alien 2":
            self.image=objs.alien2Image
        if self.id=="alien 3":
            self.image=objs.alien3Image
        self.rect.w = self.image.get_width()
        self.rect.h = self.image.get_height()
        self.add(objs.spriteGroup)
    def offScreen(self):
        return (self.rect.right < -50)
    def llHitDetect(self, ll):
        #If you hit the alien, you lose. Simple.
        if objs.deathTime:
            return
        alienRect = self.rect
        llRect = ll.rect
        if alienRect.colliderect(llRect):
            objs.deathTime=time.time()
            playSound(objs.crashAudio)
    def update(self):
        if self.justOnScreen:
            return
        if self.rect.left < objs.playArea.right:
            self.justOnScreen=True
            playSound(objs.alienAudio)
    def move(self, ll):
        #Don't move if you're off to the right!
        if self.rect.x > 800:
            return
        if self.id=="alien 1":
            #Move right and left.
            if self.moveDirection=="right":
                self.rect.x+=3
                self.moveCount[0]+=3
                if self.moveCount[0]>=200:
                    self.moveDirection="left"
                    self.moveCount[0]=0
            elif self.moveDirection=="left":
                self.rect.x-=0
                self.moveCount[0]-=0
                if self.moveCount[0]<=-100:
                    self.moveDirection="right"
                    self.moveCount[0]=0
            else:
                self.moveDirection="right"
            #Move up or down to align.
            if ll.rect.y >= self.rect.y:
                self.rect.y+=1
            if ll.rect.y <= self.rect.y:
                self.rect.y-=1
        if self.id=="alien 2":
            #Move up and down.
            if self.moveDirection=="up":
                self.rect.y+=2
                self.moveCount[1]+=2
                if self.moveCount[1]>=200:
                    self.moveDirection="down"
                    self.moveCount[1]=0
            elif self.moveDirection=="down":
                self.rect.y-=2
                self.moveCount[1]-=2
                if self.moveCount[1]<=-200:
                    self.moveDirection="up"
                    self.moveCount[1]=0
            else:
                self.moveDirection="down"
            #Move left or right
            if ll.rect.x >= self.rect.x:
                self.rect.x+=ll.scrollSpeed-1
            if ll.rect.x <= self.rect.x:
                self.rect.x-=1
        if self.id=="alien 3":
            #Close in on the player, with increasing speed.
            self.moveCount[0]+=1
            speed=0
            if self.moveCount[0]<30:
                speed=0
            elif self.moveCount[0]<60:
                speed=5
            elif self.moveCount[0]<90:
                speed=10
            else:
                speed=5

            #Move left or right
            if ll.rect.x >= self.rect.x:
                self.rect.x+=speed
            if ll.rect.x <= self.rect.x:
                self.rect.x-=speed
            #Move up or down to align.
            if ll.rect.y >= self.rect.y:
                self.rect.y+=speed
            if ll.rect.y <= self.rect.y:
                self.rect.y-=speed

class Cheez(pygame.sprite.Sprite):
    def __init__(self, args=None):
        pygame.sprite.Sprite.__init__(self)
        self.image=None
        self.rect=pygame.Rect(0,0,0,0)
        self.type="default"
        self.id="yellow cheez"
        if args:
            self.id=args[0]
            self.rect.x=args[1]
            self.rect.y=args[2]
        self.load()
    def load(self):
        if self.id=="yellow cheez":
            self.image=objs.cheezYellowImage
        if self.id=="blue cheez":
            self.image=objs.cheezBlueImage
        if self.id=="red cheez":
            self.image=objs.cheezRedImage
        self.rect.w = self.image.get_width()
        self.rect.h = self.image.get_height()
        self.add(objs.spriteGroup)
    def offScreen(self):
        return (self.rect.right < -50)

class Claw(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        self.rect=pygame.Rect(0,0,0,0)
        self.image=None
        self.scale=1.0
        self.angle=0.0
        self.status="Ready"
    def load(self):
        self.image=objs.mechClawImageReady
        self.rect.w = self.image.get_width()
        self.rect.h = self.image.get_height()
        self.add(objs.spriteGroup)
    def update(self):
        #Alter the claw if it's bigger, rotated etc.
        imageToUse=None
        if self.status=="Ready":
            imageToUse=objs.mechClawImageReady
        elif self.status=="Recover":
            imageToUse=objs.mechClawImageRecover
        elif self.status=="Locked On":
            imageToUse=objs.mechClawImageLockedOn
        self.image=pygame.transform.rotozoom(imageToUse, self.angle, self.scale)
        self.rect.w = self.image.get_width()
        self.rect.h = self.image.get_height()

class LunarLander(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        self.rect=pygame.Rect(0,0,0,0)
        self.image=None
        self.verticalVelocity = 0
        self.scrollSpeed = 2
        self.horizontalVelocity = 0
        self.pendingVerticalDown = [None, "nothing"]
        self.pendingVerticalUp = [None, "nothing"]
        self.jumpjetsActivated=None
        self.jumpjetFuelMax=0.3
        self.jumpjetFuel=0.3
        self.jumpjetFuelRechargeRate=0.3
        self.jumpStrength=1
        self.clawRange=200
        self.claw = Claw()
        self.clawTarget=[0,0]
        self.score=0
        self.clawSpeed=10
        self.clawTime=0
        self.blueCheezTime=None
        self.redCheezTime=None
        self.yellowCheezTime=None
    def load(self):
        self.image=objs.lunarLanderImage
        self.rect.w = self.image.get_width()
        self.rect.h = self.image.get_height()
        self.add(objs.spriteGroup)
        #load the claw, too!
        self.claw.load()        
    def getScrollSpeed(self):
        if self.blueCheezTime==None or time.time() - self.blueCheezTime > 5.0:
            return self.scrollSpeed
        return self.scrollSpeed + 5
    def getClawRange(self):
        if self.redCheezTime==None or time.time() - self.redCheezTime > 10.0:
            return self.clawRange
        return int(self.clawRange * 3 / 2)
    def getClawSpeed(self):
        if self.redCheezTime==None or time.time() - self.redCheezTime > 10.0:
            return self.clawSpeed
        return int(self.clawSpeed * 2)
    def getClawRecoveryTime(self):
        if self.redCheezTime==None or time.time() - self.redCheezTime > 10.0:
            return 40
        return 10
    def getJumpjetFuelMax(self):
        if self.yellowCheezTime==None or time.time() - self.yellowCheezTime > 10.0:
            return self.jumpjetFuelMax
        return self.jumpjetFuelMax * 2
    def draw(self):
        #Draw the claw range?
        if self.redCheezTime and (time.time() - self.redCheezTime > 0.0 and time.time() - self.redCheezTime < 1.0) or (time.time() - self.redCheezTime > 9.0 and time.time() - self.redCheezTime < 11.0):
            pygame.gfxdraw.circle(objs.screen, self.rect.centerx, self.rect.centery, self.getClawRange(), (192, 160-64, 160-64))
        #Draw a line to the claw
        pygame.gfxdraw.line(objs.screen, self.rect.centerx, self.rect.centery, self.claw.rect.centerx, self.claw.rect.centery, (192, 160-64, 160-64))
    def pendingMovement(self):
        #Using gravity, predict the pending movement.
        self.verticalVelocity += objs.gravityAccel
        #The LL should be at horiz position = 100. Speed up or slow down accordingly.
        if self.rect.centerx < 100:
            self.horizontalVelocity = 1
        if self.rect.centerx > 100:
            self.horizontalVelocity = -1
        if self.rect.centerx == 100:
            self.horizontalVelocity = 0
    def addPendingVertical (self, pendingAction):
        #We're taking the least amount, so.
        if pendingAction[0] < 0 and (self.pendingVerticalDown[0]==None or pendingAction[0] < self.pendingVerticalDown[0]):
            self.pendingVerticalDown = pendingAction
    def addPendingVerticalUp (self, pendingAction):
        if pendingAction[0] > 0 and (self.pendingVerticalUp[0]==None or pendingAction[0] > self.pendingVerticalUp[0]):
            self.pendingVerticalUp = pendingAction
    def move(self):
        #Move horizontally and vertically
        self.rect.x+=self.horizontalVelocity

        #Apply the pending vertical action.
        if self.pendingVerticalDown[0] == None:
            self.rect.y+=self.verticalVelocity
            #if you are jumping, continue jumping!
            if self.jumpjetsActivated:
                self.startJumping()
        else:
            self.rect.y+=self.pendingVerticalDown[0]

        #You can't go higher than the play area. Sorry.
        if self.rect.y < objs.playArea.top - 64:
            self.rect.y = objs.playArea.top - 64
            self.stopJumping()
            self.verticalVelocity=0

        if self.pendingVerticalDown[1]=="stop" and self.verticalVelocity>=0:
            self.verticalVelocity=0
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        if self.pendingVerticalDown[1]=="climb up":
            self.verticalVelocity=-1
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        if self.pendingVerticalDown[1]=="climb down":
            self.verticalVelocity=0
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        self.pendingVerticalDown = [None, "nothing"]

        #And the up direction
        if self.pendingVerticalUp[0] != None:
            self.rect.y+=self.pendingVerticalDown[0]

        if self.pendingVerticalUp[1]=="stop" and self.verticalVelocity>=0:
            self.verticalVelocity=0
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        if self.pendingVerticalUp[1]=="climb up":
            self.verticalVelocity=-1
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        if self.pendingVerticalUp[1]=="climb down":
            self.verticalVelocity=0
            self.jumpjetFuel=self.getJumpjetFuelMax()
            self.jumpjetsActivated=None
        self.pendingVerticalUp = [None, "nothing"]

        #claw movement
        #If it's locked on to something, go for it
        if self.claw.status=="Locked On":
            if self.claw.rect.centerx < self.clawTarget[0]:
                self.claw.rect.centerx+=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centerx > self.clawTarget[0]:
                self.claw.rect.centerx-=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centery < self.clawTarget[1]:
                self.claw.rect.centery+=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centery > self.clawTarget[1]:
                self.claw.rect.centery-=self.getClawSpeed()+self.getScrollSpeed()
            #If you hit the target, destroy it and recover.
            distx=self.claw.rect.centerx-self.rect.centerx
            distx*=distx
            disty=self.claw.rect.centery-self.rect.centery
            disty*=disty
            if distx + disty < 16 * 16:
                self.claw.status = "Recover"
            #If you've locked on for too long, recover.
            self.clawTime+=1 #Increment claw timer
            if self.clawTime > self.getClawRecoveryTime():
                self.claw.status = "Recover"
        else:
        #If it's out of bounds, pull it back!
            distx=self.claw.rect.centerx-self.rect.centerx
            distx*=distx
            disty=self.claw.rect.centery-self.rect.centery
            disty*=disty
            if distx + disty > self.getClawRange() * self.getClawRange():
                self.claw.status="Recover"
            if distx + disty < (self.getClawRange() * self.getClawRange())/4:
                self.claw.status="Ready"
                self.clawTime=0

        #If the claw is steady, it wants to be at a ready position.
        if self.claw.status=="Ready" or self.claw.status=="Recover":
            if self.claw.rect.centerx < self.rect.x+0:
                self.claw.rect.centerx+=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centerx > self.rect.x+20:
                self.claw.rect.centerx-=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centery < self.rect.y-20:
                self.claw.rect.centery+=self.getClawSpeed()+self.getScrollSpeed()
            if self.claw.rect.centery > self.rect.y-50:
                self.claw.rect.centery-=self.getClawSpeed()+self.getScrollSpeed()
        #update the claw's parameters
        self.setClawDrawingParameters ()

        #jet fuel regain: If you don't jump, you get fuel back.
        if self.verticalVelocity > 0:
            self.jumpjetFuel+=self.jumpjetFuelRechargeRate/30.0
            if self.jumpjetFuel>self.getJumpjetFuelMax():
                self.jumpjetFuel=self.getJumpjetFuelMax()
    def setClawDrawingParameters (self):
        #Tell the claw how big to draw itself
        if self.getClawRange() > self.clawRange:
            self.claw.scale=2.0
        else:
            self.claw.scale=1.0

        #Pointing the claw. If something is in range the claw should look at it.
        if self.claw.status=="Ready":
            nearestTarget=self.findNearestClawTarget(objs.cheezes + objs.aliens)
            if nearestTarget!=None:
                self.claw.angle=math.degrees(math.atan2(self.claw.rect.centerx-nearestTarget.rect.centerx,self.claw.rect.centery-nearestTarget.rect.centery))
            else:
                self.claw.angle=math.degrees(math.atan2(self.rect.centerx-self.claw.rect.centerx,self.rect.centery-self.claw.rect.centery))

    def findNearestClawTarget(self, validTargets):
        #Of all the targets in range, this picks the closest.
        #track all of the valid targets in the room.
        thisTarget=None
        currentDistance=None
        for target in validTargets:
            #track distances to each valid target. If it's out of range it's not in the consideration.
            distx=target.rect.centerx-self.rect.centerx
            distx*=distx
            disty=target.rect.centery-self.rect.centery
            disty*=disty
            if distx + disty < self.getClawRange() * self.getClawRange() and (not currentDistance or distx+disty < currentDistance):
                currentDistance = distx + disty
                thisTarget=target
        return thisTarget
    def useTheClaw(self, validTargets):
        #If the claw is ready, 
        if self.claw.status != "Ready":
            return
        thisTarget = self.findNearestClawTarget(validTargets)
        if thisTarget==None:
            return
        self.claw.status = "Locked On"
        self.clawTarget = [thisTarget.rect.x, thisTarget.rect.y]
    def startJumping(self):
        #Gives a boosted vertical velocity.
        if not self.jumpjetsActivated and self.jumpjetFuel > 0:
            self.verticalVelocity-=self.jumpStrength
            self.jumpjetsActivated=time.time()
        if self.jumpjetsActivated and self.jumpjetFuel > 0:
            self.verticalVelocity-=self.jumpStrength
            self.jumpjetFuel-=time.time() - self.jumpjetsActivated
            self.jumpjetsActivated=time.time()
    def stopJumping(self):
        #Stops jumping.
        self.jumpjetsActivated=None
    def clawHitDetect(self, target):
        #Did the claw hit the target?
        clawRect = self.claw.rect
        targetRect = target.rect
        typeOfTarget=None
        if clawRect.colliderect(targetRect):
            typeOfTarget = target.id
        #Reward based on the ID
            #Claw enters recover mode.
            self.claw.status="Recover"
            target.kill()
            if typeOfTarget == "yellow cheez":
                self.score+=1000
                self.yellowCheezTime=time.time()
                addLunarKid()
                playSound(objs.cheezAudio)
                return True
            if typeOfTarget == "blue cheez":
                self.score+=1500
                self.blueCheezTime=time.time()
                addLunarKid()
                playSound(objs.cheezAudio)
                return True
            if typeOfTarget == "red cheez":
                self.score+=1000
                self.redCheezTime=time.time()
                addLunarKid()
                playSound(objs.cheezAudio)
                return True
            if typeOfTarget == "alien 1":
                self.score+=1000
                return True
            if typeOfTarget == "alien 2":
                self.score+=1000
                return True
            if typeOfTarget == "alien 3":
                self.score+=1000
                return True
        return False

def horizCollide(segment1, segment2):
    #Returns true if segment1 intersects with segment2
    if segment1[0] < segment2[0] and segment1[1] < segment2[0]:
        return False
    if segment1[0] > segment2[1] and segment1[1] > segment2[1]:
        return False
    return True

def rectCollisionMatrix(rect1, rect2):
    #Returns an answer matrix of True / False values:
    #l1 < l2
    #l1 < r2
    #r1 < l2
    #r1 < r2
    #t1 < t2
    #t1 < b2
    #t2 < t2
    #t2 < b2
    answer = []
    answer.append(rect1[0] < rect2[0])
    answer.append(rect1[0] < rect2[2])
    answer.append(rect1[2] < rect2[0])
    answer.append(rect1[2] < rect2[2])
    answer.append(rect1[1] < rect2[1])
    answer.append(rect1[1] < rect2[3])
    answer.append(rect1[3] < rect2[1])
    answer.append(rect1[3] < rect2[3])
    return answer

class TerrainGround():
    def __init__(self, args=None):
        self.location = [0,0]
        self.dimensions=[30,30]
        if args and len(args) > 0:
            self.location[0]=args[0]
            self.location[1]=args[1]
            self.dimensions[0]=args[2]
            self.dimensions[1]=args[3]
    def draw(self):
        pygame.gfxdraw.box(objs.screen, pygame.Rect(self.location, self.dimensions), (200, 200, 200))
    def llHitDetectHorizontal (self, ll):
        #Did you crash into it?
        collideMatrix = rectCollisionMatrix([ll.rect.x,ll.rect.y, ll.rect.x+ll.rect.w, ll.rect.y+ll.rect.h], [self.location[0], self.location[1], self.location[0]+self.dimensions[0], self.location[1]+self.dimensions[1]])
        #We care about 0 1 2 and 6. It needs to match TTF_ __F_
        if not objs.deathTime and collideMatrix[0] and collideMatrix[1] and not collideMatrix[2] and not collideMatrix[6]:
            #compare vertical heights. If it's higher than a tolerance, then you crash.
            distance1 = ll.rect.y+ll.rect.h - self.location[1]
            distance2 = self.location[1]+self.dimensions[1] - ll.rect.y
            if distance1 >= 16 and distance2 >= 16:
                objs.deathTime=time.time()
                playSound(objs.crashAudio)
    def llHitDetectVertical(self, ll):
        collideMatrix = rectCollisionMatrix([ll.rect.x,ll.rect.y, ll.rect.x+ll.rect.w, ll.rect.y+ll.rect.h], [self.location[0], self.location[1], self.location[0]+self.dimensions[0], self.location[1]+self.dimensions[1]])
        if not objs.deathTime and ll.verticalVelocity >= 0 and collideMatrix[1] and not collideMatrix[2] and not collideMatrix[6]:
            #compare vertical heights. If it's higher than a tolerance, then you crash.
            distance1 = ll.rect.y+ll.rect.h - self.location[1]
            distance2 = self.location[1]+self.dimensions[1] - ll.rect.y
            if (distance1 >= 0 and distance1 < 16):
            #Add pending vertical landing.
                objs.ll.addPendingVertical([self.location[1]-ll.rect.y-ll.rect.h, "stop"])
            #If you jumped below a platform, provide an equal and opposite force
            if (distance2 >= 0 and distance2 < 16):
                objs.ll.addPendingVerticalUp([self.location[1]-ll.rect.y-ll.rect.h, "stop"])
    def offScreen(self):
        return (self.location[0]+self.dimensions[0] < -50)

class TerrainSlope():
    def __init__(self, args=None):
        self.slopeDirection="up" #"up" or "down"
        self.location = [0,0]
        self.size=30
        if args and len(args) > 0:
            self.slopeDirection=args[0]
            self.location[0]=args[1]
            self.location[1]=args[2]
            self.size=args[3]
    def heightAtXCoord(self, xcoord):
        #Returns the x coordinate given the level. It's a 45 degree slope, so the slope is 1 (or -1)
        if self.slopeDirection=="up":
            if xcoord < self.location[0]:
                xcoord=self.location[0]
            if xcoord > self.location[0]+self.size:
                xcoord = self.location[0]+self.size
            return -1 * xcoord + self.location[0] + self.location[1] + self.size
        if self.slopeDirection=="down":
            if xcoord < self.location[0]:
                xcoord=self.location[0]
            if xcoord > self.location[0]+self.size:
                xcoord = self.location[0]+self.size
            return xcoord + self.location[1] - self.location[0]
    def draw(self):
        if self.slopeDirection=="up":
            pygame.gfxdraw.filled_trigon(objs.screen, 
                                         self.location[0], self.location[1]+self.size,
                                         self.location[0]+self.size, self.location[1]+self.size,
                                         self.location[0]+self.size, self.location[1],
                                         (200, 200, 200))
        else:
            pygame.gfxdraw.filled_trigon(objs.screen, 
                                         self.location[0],self.location[1],
                                         self.location[0]+self.size, self.location[1]+self.size,
                                         self.location[0],self.location[1]+self.size,
                                         (200, 200, 200))
    def llHitDetectHorizontal (self, ll):
        collideMatrix = rectCollisionMatrix([ll.rect.x,ll.rect.y, ll.rect.x+ll.rect.w, ll.rect.y+ll.rect.h], [self.location[0], self.location[1], self.location[0]+self.size, self.location[1]+self.size])
        #We care about 0 1 2 and 6. It needs to match TTF_ __F_
        if not objs.deathTime and self.slopeDirection=="down" and collideMatrix[0] and collideMatrix[1] and not collideMatrix[2] and not collideMatrix[6]:
            #compare vertical heights. If it's higher than a tolerance, then you crash.
            distance1 = ll.rect.y+ll.rect.h - self.location[1]
            distance2 = self.location[1]+self.size - ll.rect.y
            if distance1 >= 16 and distance2 >= 16:
                objs.deathTime=time.time()
                playSound(objs.crashAudio)
    def llHitDetectVertical(self, ll):
        if self.slopeDirection=="up" and (horizCollide([ll.rect.x,ll.rect.x+ll.rect.w],[self.location[0],self.location[0]+self.size])) and ll.rect.y+ll.rect.h+ll.verticalVelocity>self.heightAtXCoord(ll.rect.x+ll.rect.w):
            #Add pending vertical landing.
            objs.ll.addPendingVertical([self.heightAtXCoord(ll.rect.x+ll.rect.w)-ll.rect.y-ll.rect.h, "climb up"])
        if self.slopeDirection=="down" and (horizCollide([ll.rect.x,ll.rect.x+ll.rect.w],[self.location[0],self.location[0]+self.size])) and ll.rect.y+ll.rect.h+ll.verticalVelocity>self.heightAtXCoord(ll.rect.x):
            #Add pending vertical landing.
            objs.ll.addPendingVertical([self.heightAtXCoord(ll.rect.x)-ll.rect.y-ll.rect.h, "climb down"])
    def offScreen(self):
        return (self.location[0]+self.size < -50)

objs = Session()

def mainEngine():
    #Set pics
    objs.ll.load()

    #Set the seed
    random.seed(None)

    #Create tutorial level
    objs.terrain.append(TerrainGround())
    objs.terrain[0].location=[000,490]
    objs.terrain[0].dimensions=[2000,30]
    objs.terrain.append(TerrainSlope())
    objs.terrain[1].location=[600,390]
    objs.terrain[1].size=100

    objs.cheezes.append(Cheez(["yellow cheez",400,390]))
    objs.cheezes.append(Cheez(["red cheez",600,390]))
    objs.cheezes.append(Cheez(["blue cheez",1000,20]))

    objs.aliens.append(Alien(["alien 1", 1600, 200]))

    objs.resetProgram=False

    #Give the ll red cheez for about 1 second to demonstrate the claw's range.
    objs.ll.redCheezTime = time.time()-8.5
    
    while objs.quitProgram == False and not objs.resetProgram:
        for event in pygame.event.get():
            if event.type==KEYUP and event.key==K_ESCAPE:
                objs.quitProgram=True
            if event.type==KEYDOWN and event.key==K_SPACE:
                objs.ll.startJumping()
            if event.type==KEYUP and event.key==K_SPACE:
                objs.ll.stopJumping()
            if event.type==KEYUP and event.key==K_RETURN and objs.deathMode=="Game Over Screen":
                objs.resetProgram=True
                break
            if event.type==KEYDOWN and (event.key==K_LCTRL or event.key==K_RCTRL):
                objs.ll.useTheClaw(objs.cheezes + objs.aliens)
        objs.clock.tick(30) #Wait for next clock tick
        #Clear the screen
        pygame.gfxdraw.box(objs.screen, pygame.Rect(objs.playArea) ,(16, 16, 16))
        #Draw the background?
        if objs.backgroundScrollStartTime:
            if objs.backgroundImage.get_width() > objs.playArea.w:
                if time.time() - objs.backgroundScrollStartTime > objs.backgroundScrollDuration:
                    objs.backgroundScrollPos=objs.playArea.w - objs.backgroundImage.get_width()
                else:
                    objs.backgroundScrollPos=int((objs.playArea.w - objs.backgroundImage.get_width()) * (time.time()-objs.backgroundScrollStartTime) / objs.backgroundScrollDuration)
            else:
                objs.backgroundScrollPos=objs.playArea.centerx-(objs.backgroundImage.get_width()/2)
            topPrint = objs.playArea.bottom - objs.backgroundImage.get_height()
            objs.screen.blit(objs.backgroundImage, (objs.backgroundScrollPos, topPrint))

        drawTutorialScreen()

        #Draw the terrain.
        for t in objs.terrain:
            t.draw()

        #Drawing and moving the Lunar Lander:
        objs.spriteGroup.update()
        objs.spriteGroup.draw(objs.screen)
        objs.ll.draw()#draw to screen!
        #vertical movement: calculate pending vertical
        objs.ll.pendingMovement()
        #Now ask all terrain objects to check for hit detection.
        for t in objs.terrain:
            t.llHitDetectHorizontal(objs.ll)
            t.llHitDetectVertical(objs.ll)

        #Add hit detection for aliens here
        for t in objs.aliens:
            t.llHitDetect(objs.ll)    

        #The final surface is the bottom of the play area - it should never fall off screen
        if objs.ll.rect.y+objs.ll.rect.h+objs.ll.verticalVelocity>objs.playArea.bottom and not objs.deathTime:
            #Add pending vertical landing.
            objs.ll.addPendingVertical([objs.playArea.bottom-objs.ll.rect.y-objs.ll.rect.h, "stop"])
            #You crashed and died.
            objs.deathTime=time.time()
            playSound(objs.crashAudio)

        #Scroll everything over. Remove anything that is offScreen()
        if objs.deathTime == None:
            for t in objs.terrain:
                t.location[0]-=objs.ll.getScrollSpeed()
            objs.terrain[:]=[t for t in objs.terrain if not t.offScreen()]
            for t in objs.cheezes:
                t.rect.x-=objs.ll.getScrollSpeed()
            objs.cheezes[:]=[t for t in objs.cheezes if not t.offScreen()]
            for t in objs.aliens:
                t.rect.x-=objs.ll.getScrollSpeed()
            objs.aliens[:]=[t for t in objs.aliens if not t.offScreen()]

        #Do we need to generate more terrain?
        if objs.tutorialScreen==None:
            objs.terrainCursor[0]-=objs.ll.getScrollSpeed()
        if objs.tutorialScreen==None and objs.terrainCursor[0]<=800:
            objs.ll.scrollSpeed+=1
            objs.cheezGenerator[0]+=5
            objs.alienGenerator[0]+=1
            generateNextScreen(1)

        #Increase your score based on scroll speed.
        if objs.deathTime == None:
            objs.ll.score += objs.ll.getScrollSpeed()
        #Is that a high score?
        if objs.ll.score >= objs.highScore:
            objs.highScore = objs.ll.score

        #Did the claw hit anything?
        objs.cheezes[:]=[t for t in objs.cheezes if not objs.ll.clawHitDetect(t)]
        objs.aliens[:]=[t for t in objs.aliens if not objs.ll.clawHitDetect(t)]

        #Game over yet?
        if objs.deathTime and time.time() - objs.deathTime > 1.0 and objs.deathMode == "None":
            objs.deathMode = "Game Over Screen"

        #Game Over screen
        if objs.deathMode == "Game Over Screen":
            pygame.gfxdraw.box(objs.screen, pygame.Rect(objs.playArea) ,(16, 16, 16))#Clear the screen
            objs.screen.blit(objs.defaultFont.render("GAME OVER", False, (200,200,200)), (200, 250))
            objs.screen.blit(objs.defaultFont.render("Press Enter to continue, Press ESC to quit", False, (200,200,200)), (200, 400))


        if objs.deathTime == None:
        #Move the aliens
            for t in objs.aliens:
                t.move(objs.ll)

        #Finally, move the LunarLander.
            objs.ll.move()

        #Draw the score display.
        pygame.gfxdraw.box(objs.screen, pygame.Rect(0, objs.playArea.bottom, 800, 600) ,(0, 0, 0))
        objs.screen.blit(objs.defaultFont.render("Score: "+str(objs.ll.score), False, (200,200,200)), (20, 560))
        if objs.ll.jumpjetFuel > 0:
            jumpjetLength = 200 * objs.ll.jumpjetFuel / objs.ll.getJumpjetFuelMax()
            color = (255-32,204-32,0)
            if objs.ll.getJumpjetFuelMax() > objs.ll.jumpjetFuelMax:
                color = (255,204,64)
            pygame.gfxdraw.box(objs.screen, pygame.rect.Rect((300,560), (jumpjetLength,30)), color)
            objs.screen.blit(objs.defaultFont.render("Jumpjets", False, (64,64,64)), (300, 560))
        else:
            objs.screen.blit(objs.defaultFont.render("No Fuel", False, (200,200,200)), (300, 560))
        objs.screen.blit(objs.defaultFont.render("High Score: "+str(objs.highScore), False, (200,200,200)), (550, 560))

        pygame.display.update()

def drawTutorialScreen():
    if objs.tutorialTime==None:
        objs.tutorialTime=time.time()
        objs.tutorialScroll=800
    if objs.tutorialScreen=="Title":
        guiPositions={"scroll time": 2.0, "title":[200,50], "one line summary":[100,150], "spacebar":[110,200], "ctrl":[110,250], "escape":[110,300]}
        objs.screen.blit(objs.defaultFont.render("Lunar Lander Cheez Grabber!", False, (200,200,200)), (objs.tutorialScroll+guiPositions["title"][0], guiPositions["title"][1]))
        objs.screen.blit(objs.defaultFont.render("Grab cheez. Smack aliens. Don't crash!", False, (200,200,200)), (objs.tutorialScroll+guiPositions["one line summary"][0], +guiPositions["one line summary"][1]))
        objs.screen.blit(objs.defaultFont.render("Spacebar = Jumpjets", False, (150,150,150)), (objs.tutorialScroll+guiPositions["spacebar"][0], guiPositions["spacebar"][1]))
        objs.screen.blit(objs.defaultFont.render("Ctrl = The Claw!", False, (150,150,150)), (objs.tutorialScroll+guiPositions["ctrl"][0], guiPositions["ctrl"][1]))
        objs.screen.blit(objs.defaultFont.render("Escape = Quit :(", False, (150,150,150)), (objs.tutorialScroll+guiPositions["escape"][0], guiPositions["escape"][1]))
        #Scroll the tutorial screen over.
        if time.time() - objs.tutorialTime < guiPositions["scroll time"]:
            objs.tutorialScroll=800 * (objs.tutorialTime + guiPositions["scroll time"] - time.time()) / guiPositions["scroll time"]
        if time.time() - objs.tutorialTime > guiPositions["scroll time"]+5.0:
            objs.tutorialScreen="Items"
            objs.tutorialTime=time.time()
            objs.tutorialScroll=800
    if objs.tutorialScreen=="Items":
        guiPositions={"scroll time": 1.5, "plot":[20,50], "cheez x":[50, 100, 150], "cheez y":[75,100,125], "aliens":[50,160], "good luck":[100,200]}
        objs.screen.blit(objs.defaultFont.render("Grab cheez! Lunar Children LOVE cheez!", False, (200,200,200)), (objs.tutorialScroll+guiPositions["plot"][0], guiPositions["plot"][1]))
        objs.screen.blit(objs.defaultFont.render("Yellow cheez: Better Jumpjets", False, (255,204,0)), (objs.tutorialScroll+guiPositions["cheez x"][0], guiPositions["cheez y"][0]))
        objs.screen.blit(objs.defaultFont.render("Blue cheez  : Turbo (Big point bonus)", False, (100,100,200)), (objs.tutorialScroll+guiPositions["cheez x"][1], guiPositions["cheez y"][1]))
        objs.screen.blit(objs.defaultFont.render("Red cheez   : Bigger Claw points!", False, (200,100,100)), (objs.tutorialScroll+guiPositions["cheez x"][2], guiPositions["cheez y"][2]))
        objs.screen.blit(objs.defaultFont.render("Watch out for Aliens! Bop them with your claw for 1000 points!", False, (128,128,128)), (objs.tutorialScroll+guiPositions["aliens"][0], guiPositions["aliens"][1]))
        objs.screen.blit(objs.defaultFont.render("The terrain is pretty random after this. Good luck!", False, (200,200,200)), (objs.tutorialScroll+guiPositions["good luck"][0], guiPositions["good luck"][1]))
        #Scroll the tutorial screen over.
        if time.time() - objs.tutorialTime < guiPositions["scroll time"]:
            objs.tutorialScroll=800 * (objs.tutorialTime + guiPositions["scroll time"] - time.time()) / guiPositions["scroll time"]
        if time.time() - objs.tutorialTime > guiPositions["scroll time"]+5.0:
            objs.tutorialTime=time.time()
            objs.tutorialScreen=None
            objs.ll.scrollSpeed=5
            objs.terrainCursor[0]=800
            generateNextScreen(2)
            objs.backgroundScrollStartTime=time.time()

def generateNextScreen(numScreens):
    #Generate platforms randomly.
    objs.terrainCursor[1]=550
    chosenObject=None
    sumToBeat=0
    randomLength=0
    randomStartingHeight=0
    terrainType = 0
    while objs.terrainCursor[0] < 800 + numScreens * 800:
        #Should I generate some cheez, or an alien?
        randomNumber=random.randint(0,1)
        if randomNumber==0:
            randomNumber=random.randint(0,100)
            if randomNumber<objs.cheezGenerator[0]:
                totalWeight=objs.cheezGenerator[1]+objs.cheezGenerator[2]+objs.cheezGenerator[3]
                randomNumber = random.randint(0,totalWeight)
                randomLength=objs.terrainCursor[0]+random.randint(-50,50)
                randomStartingHeight=random.randint(0,300)
                if randomNumber < objs.cheezGenerator[1]:
                    objs.cheezes.append(Cheez(["red cheez",randomLength,randomStartingHeight]))
                elif randomNumber < sum(objs.cheezGenerator[1:3]):
                    objs.cheezes.append(Cheez(["blue cheez",randomLength,randomStartingHeight]))
                else:
                    objs.cheezes.append(Cheez(["yellow cheez",randomLength,randomStartingHeight]))
        else:
            randomNumber=random.randint(0,100)
            if randomNumber<objs.alienGenerator[0]:
                totalWeight=objs.alienGenerator[1]+objs.alienGenerator[2]+objs.alienGenerator[3]
                randomNumber = random.randint(0,totalWeight)
                randomLength=objs.terrainCursor[0]+random.randint(-50,50)
                randomStartingHeight=random.randint(0,300)
                if randomNumber < objs.alienGenerator[1]:
                    objs.aliens.append(Alien(["alien 1",randomLength,objs.terrainCursor[1]+random.randint(-50,50)]))
                elif randomNumber < sum(objs.alienGenerator[1:3]):
                    objs.aliens.append(Alien(["alien 2",randomLength,objs.terrainCursor[1]+random.randint(-200,0)]))
                else:
                    objs.aliens.append(Alien(["alien 3",randomLength,objs.terrainCursor[1]+random.randint(-200,-100)]))
        #For the terrain itself, generate set pieces.
        terrainType = random.randint(0,100)
        if terrainType < 50:
            #3 ground terrain (the second could be blank)
            randomLength = random.randint(0,2)
            if randomLength==0:
                randomLength=100
            elif randomLength==1:
                randomLength=200
            else:
                randomLength=400
            objs.terrainCursor[1] = 490+random.randint(0,30)
            objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, 100]))
            objs.terrainCursor[0]+=randomLength
            #The second term may be a gap, if you desire.
            if random.randint(0,100) < 50:
                randomLength = random.randint(0,2)
                if randomLength==0:
                    randomLength=100
                elif randomLength==1:
                    randomLength=200
                else:
                    randomLength=400
                    objs.terrainCursor[1] = 490+random.randint(0,30)
                objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, 100]))
                objs.terrainCursor[0]+=randomLength
            else:
                randomLength = random.randint(0,2)
                if randomLength==0:
                    objs.terrainCursor[0]+=100
                elif randomLength==1:
                    objs.terrainCursor[0]+=150
                else:
                    objs.terrainCursor[0]+=200
            randomLength = random.randint(0,2)
            if randomLength==0:
                randomLength=100
            elif randomLength==1:
                randomLength=200
            else:
                randomLength=400
            objs.terrainCursor[1] = 490+random.randint(-16,16)
            objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, 100]))
            objs.terrainCursor[0]+=randomLength
        elif terrainType < 80:
            #slope up (hover?) slope down
            randomLength = random.randint(0,100)
            if randomLength<40:
                randomLength=100
            elif randomLength<65:
                randomLength=200
            elif randomLength<80:
                randomLength=300
            elif randomLength<90:
                randomLength=350
            else:
                randomLength=400
            #Build from the ground
            objs.terrainCursor[1] = 550-randomLength
            
            objs.terrain.append(TerrainSlope(["up",objs.terrainCursor[0], objs.terrainCursor[1], randomLength]))
            objs.terrainCursor[0]+=randomLength
            #Add a hover platform, if you want.
            if random.randint(0,1)==0:
                randomLength = random.randint(0,2)
                if randomLength==0:
                    randomLength=100
                elif randomLength==1:
                    randomLength=150
                else:
                    randomLength=200

                randomHeight = random.randint(0,100)

                if objs.terrainCursor[1] > 550:
                    objs.terrainCursor[1]=540
                elif objs.terrainCursor[1] < 0:
                    objs.terrainCursor[1]=20

                objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1]+10, randomLength, randomHeight]))
                objs.terrainCursor[0]+=randomLength
            #And the down slope
            randomLength = random.randint(0,100)
            if randomLength<40:
                randomLength=100
            elif randomLength<65:
                randomLength=200
            elif randomLength<80:
                randomLength=300
            elif randomLength<90:
                randomLength=350
            else:
                randomLength=400
            objs.terrain.append(TerrainSlope(["down",objs.terrainCursor[0], objs.terrainCursor[1]+16, randomLength]))
            objs.terrainCursor[0]+=randomLength
            objs.terrainCursor[1]+=randomLength-16
            #If you are too low, move the cursor back.
            if objs.terrainCursor[1] > objs.playArea.h:
                objs.terrainCursor[0]-=objs.terrainCursor[1]-objs.playArea.h
                objs.terrainCursor[1]=objs.playArea.h
        elif terrainType < 90:
            #slope up hover
            randomLength = random.randint(0,100)
            if randomLength<60:
                randomLength=100
            elif randomLength<80:
                randomLength=150
            else:
                randomLength=200
            objs.terrainCursor[1]-=randomLength
            #Leave room on the ceiling!
            if objs.terrainCursor[1] < 70:
                objs.terrainCursor[1]=70
            objs.terrain.append(TerrainSlope(["up",objs.terrainCursor[0], objs.terrainCursor[1], randomLength]))
            objs.terrainCursor[0]+=randomLength
            #hover platform!
            randomLength = random.randint(0,2)
            if randomLength==0:
                randomLength=50
            elif randomLength==1:
                randomLength=80
            else:
                randomLength=100

            randomHeight = random.randint(0,100)

            objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, randomHeight]))
            objs.terrainCursor[0]+=randomLength            

            objs.terrainCursor[1]+=random.randint(16,32)

            #hover platform!
            randomLength = random.randint(0,2)
            if randomLength==0:
                randomLength=50
            elif randomLength==1:
                randomLength=80
            else:
                randomLength=100

            randomHeight = random.randint(0,100)

            objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, randomHeight]))
            objs.terrainCursor[0]+=randomLength            

            objs.terrainCursor[1]+=random.randint(-16,16)
            
            if objs.terrainCursor[1] > 550:
                objs.terrainCursor[1]=540
            elif objs.terrainCursor[1] < 0:
                objs.terrainCursor[1]=20
        else:
            #slope up gap wall
            randomLength = random.randint(0,100)
            if randomLength<60:
                randomLength=100
            elif randomLength<80:
                randomLength=150
            elif randomLength<90:
                randomLength=200
            else:
                randomLength=250
            #Build from the ground
            objs.terrainCursor[1] = 550-randomLength
            
            objs.terrain.append(TerrainSlope(["up",objs.terrainCursor[0], objs.terrainCursor[1], randomLength]))
            objs.terrainCursor[0]+=randomLength

            #And now, a gap.
            randomLength = random.randint(0,100)
            if randomLength<60:
                randomLength=100
            elif randomLength<80:
                randomLength=150
            elif randomLength<90:
                randomLength=200
            else:
                randomLength=250
            objs.terrainCursor[0]+=randomLength

            #Finally, a wall. It should be around the height of the slope.
            randomLength=int(randomLength + random.randint(-20,20))
            objs.terrainCursor[1] = 550-randomLength

            objs.terrain.append(TerrainGround([objs.terrainCursor[0], objs.terrainCursor[1], randomLength, objs.playArea.bottom]))
            objs.terrainCursor[0]+=randomLength            

def addLunarKid():
    pos = (random.randint(objs.backgroundScrollPos,objs.backgroundScrollPos+objs.playArea.w), random.randint(0,objs.playArea.h/2))
    objs.backgroundImage.blit(objs.lunarKidImage, pos)

def playSound(zound):
    zound.play()

def main():
    #Resolution, and timer
    #The screen
    objs.screen = pygame.display.set_mode((800,600))
    
    #Game timer
    objs.clock = pygame.time.Clock()

    #mixer
    pygame.mixer.init()

    #Load fonts
    pygame.font.init()
    objs.defaultFont = pygame.font.Font("resources/babyUniverseItalic.ttf", 20)

    #Load graphics
    objs.lunarLanderImage=pygame.image.load("resources/MoonRoverFinal.png").convert_alpha()
    objs.alien1Image=pygame.image.load("resources/Alien1Final.png").convert_alpha()
    objs.alien2Image=pygame.image.load("resources/Alien2Final.png").convert_alpha()
    objs.alien3Image=pygame.image.load("resources/Alien3Final.png").convert_alpha()
    objs.cheezYellowImage=pygame.image.load("resources/Cheese3Final.png").convert_alpha()
    objs.cheezBlueImage=pygame.image.load("resources/Cheese2Final.png").convert_alpha()
    objs.cheezRedImage=pygame.image.load("resources/Cheese1Final.png").convert_alpha()
    objs.mechClawImageReady=pygame.image.load("resources/MechanicalClawReady.png").convert_alpha()
    objs.mechClawImageRecover=pygame.image.load("resources/MechanicalClawRecover.png").convert_alpha()
    objs.mechClawImageLockedOn=pygame.image.load("resources/MechanicalClawLockedOn.png").convert_alpha()
    objs.lunarKidImage=pygame.image.load("resources/lunarKid.png").convert_alpha()
    objs.alienAudio=pygame.mixer.Sound("resources/alienAudio.ogg")
    objs.cheezAudio=pygame.mixer.Sound("resources/cheezAudio.ogg")
    objs.crashAudio=pygame.mixer.Sound("resources/crashAudio.ogg")

    while not objs.quitProgram:
        objs.reset()
        mainEngine()

if __name__ == '__main__': main()
